%{

/******************************************************************************
* MODULE     : mm_parser.y / mm_parser.cc
* DESCRIPTION: The mathemagix parser
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <mm_symbs.hh>
#include <mm_lexer.hh>
#include <mm_parser.hh>
#include <error.hh>

#define malloc GC_MALLOC
#define free(x)

#define YYSTYPE generic
#define yyparse mm_parse
#define yylex mm_lex
#define yyerror mm_error
#define yylval mm_lval
#define yychar mm_char
#define yydebug mm_debug
#define yynerrs mm_nerrs

static generic mm_parse_result;
static list    mm_parse_warnings;
static list    mm_parse_errors;
static void mm_warning (char* message, generic where);
static void mm_error (char* message);

/******************************************************************************
* Generation of expressions
******************************************************************************/

static generic if_then (generic g_if, generic g_cond, generic g_arg);
static generic sequence (generic g_op, generic g_arg);

%}

/******************************************************************************
* One shift/reduce conflict expected (dangling else)
******************************************************************************/

%expect 1

/******************************************************************************
* The tokens
******************************************************************************/

%token IDENTIFIER NUMBER STRING

%token ADD BREAK CASE CATCH CONTINUE DO ELSE EXPORT EXTEND FOR FROM
%token GENERATE HAS HIDDEN HOLDS IF IMPORT IN INFIX INHERIT INTERACTIVE
%token LOOP PREFIX POSTFIX RAISE RETURN STEP THEN TRY UNTIL VALUE
%token WHILE WITH YIELD MACRO

%token ABSTRACT CONCRETE CONSTANT DIRECT EVOLUTIVE EXTERN INDIRECT
%token INLINE INTERN LOCKED METHOD PRIVATE PROTECTED OUTLINE
%token PACKED PUBLIC UNPACKED VARIABLE

%token DEFINE ASSIGN PLUS_ASSIGN MINUS_ASSIGN TIMES_ASSIGN OVER_ASSIGN
%token TYPE
%token INTO MAPSTO
%token LEFT_FLUX RIGHT_FLUX
%token FORALL EXISTS
%token WHERE
%token IMPLIES EQUIVALENT
%token SEQOR OR XOR
%token SEQAND AND
%token NOT
%token EQUAL LESS LEQ GREATER GEQ
%token NOT_EQUAL NOT_LESS NOT_LEQ NOT_GREATER NOT_GEQ
%token CAST
%token RANGE
%token PLUS MINUS OPLUS OMINUS
%token TIMES OVER OTIMES OOVER COMPOSE
%token SIGNATURE
%token POWER
%token INC DEC SIZE
%token ACCESS
%token PRIME
%token THIS FOREIGN

%%

Program : Dl                     { mm_parse_result= sequence (MM_DECLARE,$1); }
        | INTERACTIVE E0         { mm_parse_result= $2; }
        | INTERACTIVE E0 Sep     { mm_parse_result= $2; }
        ;

/******************************************************************************
* Declarations in main program or inside with/add blocks
******************************************************************************/

D0      : error Sep              { yyerrok; $$ = MM_ERROR; }
        | IMPORT IE Sep          { $$ = extend (list ($1, $2), $3); }
        | EXPORT IE Sep          { $$ = extend (list ($1, $2), $3); }
        | Var TYPE F0 Sep        { $$ = extend (list ($2, $1, $3), $4); }
        | Var TYPE WithAdd       { $$ = list ($2, $1, $3); }
        | TypeDecl DEFINE EC     { $$ = list ($2, $1, $3); }
        | TypeDecl ASSIGN EC     { $$ = list ($2, $1, $3); }
        ;

IE      : Es                     { $$ = $1; }
        | Es FROM Es             { $$ = list ($2, $1, $3); }
        ;

Var     : Id                     { $$ = $1; }
        | Prefix                 { $$ = $1; }
        | OnlyPost               { $$ = $1; }
        | Infix                  { $$ = $1; }
        | Minus                  { $$ = $1; }
        | '(' Vars ')'           { $$ = cons (symbol ("tuple!", $1), $2);
	                           $$ = extend ($$, $3); }
        | Var '(' Es ')'         { $$ = extend (cons ($1, $3), $4); }
        ;

Vars    : Var                    { $$ = list ($1); }
        | Var ',' Vars           { $$ = cons ($1, $3); }
        ;

TypeDecl: Var TYPE F0            { $$ = list ($2, $1, $3); }
        | Var TYPE WithAdd       { $$ = list ($2, $1, $3); }
        ;

Dl      :                        { $$ = list (); }
        | D0 Dl                  { $$ = cons ($1, $2); }
        | Mod Dl                 { $$ = apply_all ($1, as_list ($2)); }
        | Mod '{' Dl '}'         { $$ = apply_all ($1, as_list ($3));
	                           $$ = extend ($$, $4); }
        ;

Mod     : DeclMod                { $$ = list (MM_MODIFY, $1); }
        | TypeMod                { $$ = list (MM_MODIFY, $1); }
        | IF E0 THEN             { $$ = list (MM_MODIFY, $1, $2);
	                           $$ = extend ($$, $3); }
        | Quantf Qvars           { $$ = cons (MM_MODIFY, $1,
					      as_list ($2)); }
        ;

Db2     : '{' Dl '}'             { $$ = extend (extend ($2, $1), $3); }
        | Mod Db2                { $$ = apply_all ($1, as_list ($2)); }
        ;

Db      : Db2                    { $$ = sequence (MM_DECLARE, $1); }
        ;

/******************************************************************************
* Commands
******************************************************************************/

C0      : error Sep              { yyerrok; $$ = MM_ERROR; }
        | E1 Sep                 { $$ = extend ($1, $2); }
        | Cb                     { $$ = $1; }
        | IF E0 THEN C0          { $$ = if_then ($1, $2, $4); }
        | IF E0 THEN C0 ELSE C0  { $$ = list ($1, $2, $4, $6); }
        | Conds LOOP C0          { $$ = list ($2, $1, $3); }
        | GENERATE C0            { $$ = list ($1, $2); }
        | CASE E0 IN Cases       { $$ = list ($1, $2, $4); }
        | F0 DEFINE EC           { $$ = list ($2, $1, $3); }
        | F0 ASSIGN EC           { $$ = list ($2, $1, $3); }
        | F0 PLUS_ASSIGN EC      { $$ = list ($2, $1, $3); }
        | F0 MINUS_ASSIGN EC     { $$ = list ($2, $1, $3); }
        | F0 TIMES_ASSIGN EC     { $$ = list ($2, $1, $3); }
        | F0 OVER_ASSIGN EC      { $$ = list ($2, $1, $3); }
        | RETURN EC              { $$ = list ($1, $2); }
        | BREAK EC               { $$ = list ($1, $2); }
        | YIELD EC               { $$ = list ($1, $2); }
        | RETURN Sep             { $$ = extend (list ($1), $2); }
        | CONTINUE Sep           { $$ = extend (list ($1), $2); }
        | BREAK Sep              { $$ = extend (list ($1), $2); }
        | YIELD Sep              { $$ = extend (list ($1), $2); }
        ;

Conds   :                        { $$ = list (); }
        | FOR E0 Conds           { $$ = cons (list ($1, $2), $3); }
        | FOR E0 IN E0 Conds     { $$ = cons (list ($1, $2, $4), $5); }
        | WHILE E0 Conds         { $$ = cons (list ($1, $2), $3); }
        | UNTIL E0 Conds         { $$ = cons (list ($1, $2), $3); }
        | STEP E0 Conds          { $$ = cons (list ($1, $2), $3); }
        ;

Cases   : '{' Casel '}'          { $$ = extend (extend ($2, $1), $3); }
        ;

Casel   :                        { $$ = list (); }
        | ELSE C0                { $$ = list (generic (list ($1, $2))); }
        | Among DO C0 Casel      { $$ = cons (list ($1, $3), $4); }
        ;

Among   : T0                     { $$ = list ($1); }
        | T0 ',' Among           { $$ = cons ($1, $3); }
        ;

WithAdd : F0 WITH Db             { $$ = list ($2, $1, $3); }
        | ADD Db                 { $$ = list ($1, $2); }
        ;

Sep     : ';'                    { $$ = list (); }
        | ';' Sep                { $$ = list (); }
        ;

Cw      : C0                     { $$ = $1; }
        | WithAdd                { $$ = $1; }
        ;

Cd      : C0                     { $$ = $1; }
        | IMPORT IE Sep          { $$ = extend (list ($1, $2), $3); }
        | F0 TYPE F0 Sep         { $$ = extend (list ($2, $1, $3), $4); }
        | F0 TYPE F0 DEFINE Cw   { $$ = list ($4, list ($2, $1, $3), $5); }
        | F0 TYPE F0 ASSIGN Cw   { $$ = list ($4, list ($2, $1, $3), $5); }
        | DeclMod D0             { $$ = list ($1, $2); }
        ;

Cl      :                        { $$ = list (); }
        | Cd Cl                  { $$ = cons ($1, $2); }
        ;

Cb      : '{' Cl '}'             { $$ = sequence (symbol ("begin!", $1), $2);
                                   $$ = extend ($$, $3); }
        ;

/******************************************************************************
* Expressions followed by a separator
******************************************************************************/

EC      : E1 Sep                 { $$ = extend ($1, $2); }
        | WithAdd                { $$ = $1; }
        | Cb                     { $$ = $1; }
        | IF E0 THEN E0 ELSE EC  { $$ = list ($1, $2, $4, $6); }
        | GENERATE EC            { $$ = list ($1, $2); }
        | CASE E0 FROM Cases     { $$ = list ($1, $2, $4); }
        | Conds LOOP EC          { $$ = list ($2, $1, $3); }
        | F0 TYPE F0 Sep         { $$ = list ($2, $1, $3); }
        | F0 TYPE F0 DEFINE EC   { $$ = list ($4, list ($2, $1, $3), $5); }
        | F0 TYPE F0 ASSIGN EC   { $$ = list ($4, list ($2, $1, $3), $5); }
        | F0 DEFINE EC           { $$ = list ($2, $1, $3); }
        | F0 ASSIGN EC           { $$ = list ($2, $1, $3); }
        | F0 PLUS_ASSIGN EC      { $$ = list ($2, $1, $3); }
        | F0 MINUS_ASSIGN EC     { $$ = list ($2, $1, $3); }
        | F0 TIMES_ASSIGN EC     { $$ = list ($2, $1, $3); }
        | F0 OVER_ASSIGN EC      { $$ = list ($2, $1, $3); }
        ;

/******************************************************************************
* Expressions
******************************************************************************/

E0      : E1                     { $$ = $1; }
        | WithAdd                { $$ = $1; }
        | Cb                     { $$ = $1; }
        | IF E0 THEN E0 ELSE E0  { $$ = list ($1, $2, $4, $6); }
        | GENERATE E0            { $$ = list ($1, $2); }
        | CASE E0 FROM Cases     { $$ = list ($1, $2, $4); }
        | Conds LOOP E0          { $$ = list ($2, $1, $3); }
        | F0 TYPE F0             { $$ = list ($2, $1, $3); }
        | F0 TYPE F0 DEFINE E0   { $$ = list ($4, list ($2, $1, $3), $5); }
        | F0 TYPE F0 ASSIGN E0   { $$ = list ($4, list ($2, $1, $3), $5); }
        | F0 DEFINE E0           { $$ = list ($2, $1, $3); }
        | F0 ASSIGN E0           { $$ = list ($2, $1, $3); }
        | F0 PLUS_ASSIGN E0      { $$ = list ($2, $1, $3); }
        | F0 MINUS_ASSIGN E0     { $$ = list ($2, $1, $3); }
        | F0 TIMES_ASSIGN E0     { $$ = list ($2, $1, $3); }
        | F0 OVER_ASSIGN E0      { $$ = list ($2, $1, $3); }
        ;

E1      : F0                     { $$ = $1; }
        | Lflux LEFT_FLUX F0     { $$ = list ($2, $1, $3); }
        | F0 RIGHT_FLUX Rflux    { $$ = list ($2, $1, $3); }
        ;

Lflux   : F0                     { $$ = $1; }
        | Lflux LEFT_FLUX F0     { $$ = list ($2, $1, $3); }
        ;

Rflux   : F0                     { $$ = $1; }
        | F0 RIGHT_FLUX Rflux    { $$ = list ($2, $1, $3); }
        ;

Es2     :                        { $$= list (); }
        | ',' E0 Es2             { $$= cons ($2, $3); }
        ;

Es      :                        { $$= list (); }
        | E0 Es2                 { $$= cons ($1, $2); }
        ;

/******************************************************************************
* Logical formulas
******************************************************************************/

Qvars   : '(' Es ')'             { $$ = $2; }
        ;

F0      : Quantf Qvars F0        { $$ = cons ($1, as_list ($2)) * list ($3); }
        | F1                     { $$ = $1; }
        ;

F1      : F2                     { $$ = $1; }
        | F2 IMPLIES F2          { $$ = list ($2, $1, $3); }
        | F2 EQUIVALENT F2       { $$ = list ($2, $1, $3); }
        ;

F2      : F3                     { $$ = $1; }
        | F3 SEQOR Seqors        { $$ = list ($2, $1, $3); }
        | F3 OR Ors              { $$ = list ($2, $1, $3); }
        | F3 XOR Xors            { $$ = list ($2, $1, $3); }
        ;

Seqors  : F3                     { $$ = $1; }
        | F3 SEQOR Seqors        { $$ = list ($2, $1, $3); }
        ;

Ors     : F3                     { $$ = $1; }
        | F3 OR Ors              { $$ = list ($2, $1, $3); }
        ;

Xors    : F3                     { $$ = $1; }
        | F3 XOR Xors            { $$ = list ($2, $1, $3); }
        ;

F3      : F4                     { $$ = $1; }
        | F4 SEQAND Seqands      { $$ = list ($2, $1, $3); }
        | F4 AND Ands            { $$ = list ($2, $1, $3); }
        ;

Seqands : F4                     { $$ = $1; }
        | F4 SEQAND Seqands      { $$ = list ($2, $1, $3); }
        ;

Ands    : F4                     { $$ = $1; }
        | F4 AND Ands            { $$ = list ($2, $1, $3); }
        ;

F4      : F5                     { $$ = $1; }
        | NOT F5                 { $$ = list ($1, $2); }
        ;

F5      : T0                     { $$ = $1; }
        | T0 EQUAL T0            { $$ = list ($2, $1, $3); }
        | T0 LESS T0             { $$ = list ($2, $1, $3); }
        | T0 LEQ T0              { $$ = list ($2, $1, $3); }
        | T0 GREATER T0          { $$ = list ($2, $1, $3); }
        | T0 GEQ T0              { $$ = list ($2, $1, $3); }
        | T0 NOT_EQUAL T0        { $$ = list ($2, $1, $3); }
        | T0 NOT_LESS T0         { $$ = list ($2, $1, $3); }
        | T0 NOT_LEQ T0          { $$ = list ($2, $1, $3); }
        | T0 NOT_GREATER T0      { $$ = list ($2, $1, $3); }
        | T0 NOT_GEQ T0          { $$ = list ($2, $1, $3); }
        | T0 HAS T0              { $$ = list ($2, $1, $3); }
        | T0 HOLDS T0            { $$ = list ($2, $1, $3); }
        ;

/******************************************************************************
* Terms
******************************************************************************/

T0      : T1                     { $$ = $1; }
        | T0 CAST T1             { $$ = list ($2, $1, $3); }
        ;

T1      : T2                     { $$ = $1; }
        | T1 INTO T2             { $$ = list ($2, $1, $3); }
        ;

T2      : T3                     { $$ = $1; }
        | T3 RANGE T3            { $$ = list ($2, $1, $3); }
        ;

T3      : T4                     { $$ = $1; }
        | T3 PLUS T4             { $$ = list ($2, $1, $3); }
        | T3 MINUS T4            { $$ = list ($2, $1, $3); }
        | T3 OPLUS T4            { $$ = list ($2, $1, $3); }
        | T3 OMINUS T4           { $$ = list ($2, $1, $3); }
        ;

T4      : T5                     { $$ = $1; }
        | T4 TIMES T5            { $$ = list ($2, $1, $3); }
        | T4 OVER T5             { $$ = list ($2, $1, $3); }
        | T4 OTIMES T5           { $$ = list ($2, $1, $3); }
        | T4 OOVER T5            { $$ = list ($2, $1, $3); }
        | T4 COMPOSE T5          { $$ = list ($2, $1, $3); }
        ;

T5      : R0                     { $$ = $1; }
        | MINUS T5               { $$ = list ($1, $2); }
        | INC T5                 { $$ = list ($1, $2); }
        | DEC T5                 { $$ = list ($1, $2); }
        | SIZE T5                { $$ = list ($1, $2); }
        ;

/******************************************************************************
* Radical terms (twice, because of 'sin x' like function application)
******************************************************************************/

R0      : R1                     { $$ = $1; }
        | R1 S0                  { $$ = list ($1, $2); }
        | TypeMod S0             { $$ = list ($1, $2); }
        | SIGNATURE S0           { $$ = list ($1, $2); }
        ;

R1      : R2                     { $$ = $1; }
        | R2 POWER R1            { $$ = list ($2, $1, $3); }
        ;

R2      : R3                     { $$ = $1; }
        | R2 INC                 { $$ = list (symbol (".++", $2), $1); }
        | R2 DEC                 { $$ = list (symbol (".--", $2), $1); }
        | R2 NOT                 { $$ = list (symbol (".!", $2), $1); }
        ;

R3      : R4                     { $$ = $1; }
        | R3 ACCESS R4           { $$ = list ($2, $1, $3); }
        | R3 '(' Es ')'          { $$ = extend (cons ($1, $3), $4); }
        | R3 '[' Es ']'          { $$ = cons (symbol (".[]", $2), $1, $3);
	                           $$ = extend ($$, $4); }
        ;

R4      : R5                     { $$ = $1; }
        | R4 PRIME               { $$ = list ($2, $1); }
        ;

R5      : S5                     { $$ = $1; }
        | '(' ')'                { $$ = list (symbol ("tuple!", $1));
	                           $$ = extend ($$, $2); }
        | '(' E0 ')'             { $$ = $2; }
        | '(' E0 ',' E0 Es2 ')'  { $$ = cons (symbol ("tuple!", $1), $2,$4,$5);
	                           $$ = extend ($$, $6); }
        | '[' Es ']'             { $$ = cons (symbol ("[]", $1), $2);
	                           $$ = extend ($$, $3); }
        ;

S0      : S1                     { $$ = $1; }
        | S1 S0                  { $$ = list ($1, $2); }
        | TypeMod S0             { $$ = list ($1, $2); }
        | SIGNATURE S0           { $$ = list ($1, $2); }
        ;

S1      : S2                     { $$ = $1; }
        | S2 POWER S1            { $$ = list ($2, $1, $3); }
        ;

S2      : S3                     { $$ = $1; }
        | S2 INC                 { $$ = list (symbol (".++", $2), $1); }
        | S2 DEC                 { $$ = list (symbol (".--", $2), $1); }
        | S2 NOT                 { $$ = list (symbol (".!", $2), $1); }
        ;

S3      : S4                     { $$ = $1; }
        | S3 ACCESS S4           { $$ = list ($2, $1, $3); }
        | S3 '(' Es ')'          { $$ = extend (cons ($1, $3), $4); }
        | S3 '[' Es ']'          { $$ = cons (symbol (".[]", $2), $1, $3);
	                           $$ = extend ($$, $4); }
        ;

S4      : S5                     { $$ = $1; }
        | S4 PRIME               { $$ = list ($2, $1); }
        ;

S5      : Id                     { $$ = $1; }
        | NUMBER                 { $$ = $1; }
        | STRING                 { $$ = $1; }
        | FOREIGN Id IN Id       { $$ = list ($1, $2, $4); }
        | FOREIGN STRING IN Id   { $$ = list ($1, $2, $4); }
        ;

/******************************************************************************
* Modifiers
******************************************************************************/

DeclMod : EVOLUTIVE              { $$ = $1; }
        | EXTERN                 { $$ = $1; }
        | INLINE                 { $$ = $1; }
        | INTERN                 { $$ = $1; }
        | LOCKED                 { $$ = $1; }
        | METHOD                 { $$ = $1; }
        | PRIVATE                { $$ = $1; }
        | PROTECTED              { $$ = $1; }
        | OUTLINE                { $$ = $1; }
        | PUBLIC                 { $$ = $1; }
        ;

TypeMod : ABSTRACT               { $$ = $1; }
        | CONCRETE               { $$ = $1; }
        | CONSTANT               { $$ = $1; }
        | DIRECT                 { $$ = $1; }
        | INDIRECT               { $$ = $1; }
        | PACKED                 { $$ = $1; }
        | UNPACKED               { $$ = $1; }
        | VARIABLE               { $$ = $1; }
        ;

/******************************************************************************
* Identifiers
******************************************************************************/

Id      : IDENTIFIER             { $$ = $1; }
        | PREFIX Prefix          { $$ = $2; }
        | PREFIX Minus           { $$ = $2; }
        | POSTFIX Postfix        { $$ = $2; }
        | INFIX Infix            { $$ = $2; }
        | INFIX Minus            { $$ = $2; }
        | THIS                   { $$ = $1; }
        ;

Quantf  : FORALL                 { $$ = $1; }
        | EXISTS                 { $$ = $1; }
        ;

Prefix  : SIZE                   { $$ = $1; }
        | NOT                    { $$ = $1; }
        | INC                    { $$ = $1; }
        | DEC                    { $$ = $1; }
        ;

OnlyPost: PRIME                  { $$ = $1; }
        ;

Postfix : OnlyPost               { $$ = $1; }
        | NOT                    { $$ = symbol (".!", $1); }
        | INC                    { $$ = symbol (".++", $1); }
        | DEC                    { $$ = symbol (".--", $1); }
        ;

Infix   : LEFT_FLUX              { $$ = $1; }
        | RIGHT_FLUX             { $$ = $1; }
        | SEQOR                  { $$ = $1; }
        | OR                     { $$ = $1; }
        | XOR                    { $$ = $1; }
        | SEQAND                 { $$ = $1; }
        | AND                    { $$ = $1; }
        | IMPLIES                { $$ = $1; }
        | EQUIVALENT             { $$ = $1; }
        | EQUAL                  { $$ = $1; }
        | LESS                   { $$ = $1; }
        | LEQ                    { $$ = $1; }
        | GREATER                { $$ = $1; }
        | GEQ                    { $$ = $1; }
        | NOT_EQUAL              { $$ = $1; }
        | NOT_LESS               { $$ = $1; }
        | NOT_LEQ                { $$ = $1; }
        | NOT_GREATER            { $$ = $1; }
        | NOT_GEQ                { $$ = $1; }
        | RANGE                  { $$ = $1; }
        | PLUS                   { $$ = $1; }
        | OPLUS                  { $$ = $1; }
        | TIMES                  { $$ = $1; }
        | OVER                   { $$ = $1; }
        | OTIMES                 { $$ = $1; }
        | OOVER                  { $$ = $1; }
        | COMPOSE                { $$ = $1; }
        | POWER                  { $$ = $1; }
        ;

Minus   : MINUS                  { $$ = $1; }
        | OMINUS                 { $$ = $1; }
        ;

%%

/******************************************************************************
* Finalize certain constructions
******************************************************************************/

static generic
if_then (generic g_if, generic g_cond, generic g_arg) {
  generic g= list (g_if, g_cond, g_arg);
  if ((N (g_arg) == 4) && (car (g_arg) == MM_IF)) {
    generic g_else = g_arg[3];
    mm_warning ("dangling else", g_else);
  }
  return g;
}

static int
status (generic g) {
  if ((N(g) == 2) && (N(g[0]) >= 2) && (g[0][0] == MM_MODIFY))
    return status (g[1]);
  if ((N(g) == 3) && (g[0] == MM_TYPE)) return 1;
  if ((N(g) == 3) && ((g[0] == MM_DEFINE) || (g[0] == MM_ASSIGN)))
    if ((N(g[1]) == 3) && (g[1][0] == MM_TYPE)) return 1;
  if ((N(g) == 2) && (g[0] == MM_IMPORT)) return 2;
  if ((N(g) == 2) && (g[0] == MM_EXPORT)) return 2;
  return 0;
}

static generic
rewrite (generic g) {
  if ((N(g) == 2) && (N(g[0]) >= 2) && (g[0][0] == MM_MODIFY))
    return list ((N(g[0])==2 ? g[0][1]: cdr (g[0])), rewrite (g[1]));
  else return g;
}

static generic
sequence (generic g_op, generic g_arg) {
  if (nil (g_arg)) return list (g_op);
  if (car (g_arg) == MM_ERROR)
    return sequence (g_op, cdr (g_arg));

  generic item= rewrite (car (g_arg));
  generic next= sequence (g_op, cdr (g_arg));
  generic head= car (next);
  generic tail= cdr (next);

  switch (status (car (g_arg))) {
  case 0:
    if (head == MM_BEGIN)
      return cons (symbol ("begin!", g_op), item, tail);
    return list (symbol ("begin!", g_op), item, next);
  case 1:
    if (head == MM_DECLARE)
      return cons (symbol ("declare!", g_op), item, tail);
    if (head == MM_LOCAL)
      return cons (symbol ("local!", g_op), item, tail);
    return list (symbol ("local!", g_op), item, next);
  case 2:
    if (head == MM_MIGRATE)
      return cons (symbol ("migrate!", g_op), item, tail);
    return list (symbol ("migrate!", g_op), item, next);
  }

  fatal_error ("unexpected situation", "sequence", "mm_parser.y");
  exit (1);
}

/******************************************************************************
* Exported parsing functions
******************************************************************************/

void
mm_warning (char* s, generic where) {
  generic war= std_warning (mm_string (s), where);
  mm_parse_errors= cons (war, mm_parse_errors);
}

void
mm_error (char* s) {
  int pos1= mm_lex_origin + mm_lex_prev;
  int pos2= mm_lex_origin + mm_lex_pos;
  generic err= std_error (mm_string (s), symbol ("", pos1, pos2));
  mm_parse_errors= cons (err, mm_parse_errors);
}

generic
mm_parse (mm_string s, list& errors) {
  start_lexer (s);
  mm_parse_result= generic ();
  mm_parse_errors= list ();
  mm_parse ();
  errors= mm_parse_errors;
  return mm_parse_result;
}
