
/******************************************************************************
* MODULE     : mm_string.cc
* DESCRIPTION: Fixed size mm_strings. Zero characters can be part of mm_string.
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <mm_string.hh>
#include <string.h>

/******************************************************************************
* The representation type
******************************************************************************/

static inline int
round_length (int n) {
  n=(n+3)&(0xfffffffc);
  if (n<24) return n;
  register int i=32;
  while (n>i) i<<=1;
  return i;
}

mm_string_rep::mm_string_rep (int n2):
  n(n2), a ((n==0) ? ((char*) NULL): new char[round_length(n)]) {}

void
mm_string_rep::resize (int n2) {
  register int l= round_length (n);
  register int m= round_length (n2);
  if (l!=m) {
    if (n2!=0) {
      register int i, k=(n2<n?n2:n);
      register char* b= new char[m];
      for (i=0; i<k; i++) b[i]=a[i];
      if (l!=0) delete[] a;
      a=b;
    }
    else if (l!=0) delete[] a;
  }
  n=n2;
}

/******************************************************************************
* Required functions for generic structures
******************************************************************************/

int
mm_string_rep::Type () {
  return MM_STRING_TYPE;
}

int
mm_string_rep::Hash () {
  register int i, h=0;
  for (i=0; i<n; i++) {
    h=(h<<9)+(h>>23);
    h=h+((int) a[i]);
  }
  return h;
}

bool
mm_string_rep::Equal (generic g) {
  if (gentype (g) != MM_STRING_TYPE) return false;
  return mm_string (this) == as_mm_string (g);
}

generic
mm_string_rep::Copy () {
  return copy (mm_string (this));
}

void
mm_string_rep::Print (ostream& out) {
  out << mm_string (this);
}

/******************************************************************************
* Constructors and convertors for mm_strings
******************************************************************************/

mm_string::mm_string (char c) {
  rep= new mm_string_rep(1);
  rep->a[0]=c;
}

mm_string::mm_string (char *a) {
  register int i, n=strlen(a);
  rep= new mm_string_rep(n);
  for (i=0; i<n; i++)
    rep->a[i]=a[i];
}

mm_string::mm_string (char* a, int n) {
  register int i;
  rep= new mm_string_rep(n);
  for (i=0; i<n; i++)
    rep->a[i]=a[i];
}

/******************************************************************************
* STring comparisons
******************************************************************************/

bool
mm_string::operator == (char* s) {
  register int i, n= rep->n;
  register char* S= rep->a;
  for (i=0; i<n; i++) {
    if (s[i]!=S[i]) return false;
    if (s[i]=='\0') return false;
  }
  return (s[i]=='\0');
}

bool
mm_string::operator != (char* s) {
  register int i, n= rep->n;
  register char* S= rep->a;
  for (i=0; i<n; i++) {
    if (s[i]!=S[i]) return true;
    if (s[i]=='\0') return true;
  }
  return (s[i]!='\0');
}

bool
mm_string::operator == (mm_string s) {
  register int i;
  if (rep->n!=s.rep->n) return false;
  for (i=0; i<rep->n; i++)
    if (rep->a[i]!=s.rep->a[i]) return false;
  return true;
}

bool
mm_string::operator != (mm_string s) {
  register int i;
  if (rep->n!=s.rep->n) return true;
  for (i=0; i<rep->n; i++)
    if (rep->a[i]!=s.rep->a[i]) return true;
  return false;
}

bool
operator < (mm_string s1, mm_string s2) {
  register int i;
  for (i=0; i<N(s1); i++) {
    if (i>=N(s2)) return false;
    if (s1[i]<s2[i]) return true;
    if (s2[i]<s1[i]) return false;
  }
  return false;
}

bool
operator <= (mm_string s1, mm_string s2) {
  register int i;
  for (i=0; i<N(s1); i++) {
    if (i>=N(s2)) return false;
    if (s1[i]<s2[i]) return true;
    if (s2[i]<s1[i]) return false;
  }
  return true;
}

/******************************************************************************
* Other operations on mm_strings
******************************************************************************/

mm_string
mm_string::operator () (int i1, int i2) {
  register int i;
  mm_string r (i2-i1);
  for (i=i1; i<i2; i++) r[i-i1]=rep->a[i];
  if (rep->begin != -1) {
    r->begin= rep->begin + i1;
    r->end  = rep->begin + i2;
  }
  return r;
}

ostream&
operator << (ostream& out, mm_string s) {
  int i, n=N(s);
  out << "\042";
  for (i=0; i<n; i++) out << s[i];
  out << "\042";
  return out;
}

mm_string
copy (mm_string s) {
  register int i, n=N(s);
  mm_string r (n);
  for (i=0; i<n; i++) r[i]=s[i];
  if (s->begin != -1) {
    r->begin= s->begin;
    r->end  = s->end;
  }
  return r;
}

static void
print_int (mm_string& s, int i) {
  if (i<10) s << ((char) (((int) '0') + i));
  else {
    print_int (s, i/10);
    s << ((char) (((int) '0') + (i%10)));
  }
}

mm_string
as_mm_string (int i) {
  mm_string s ("");
  if (i < 0) { s << "-"; i=-i; }
  print_int (s, i);
  return s;
}

char*
as_charp (mm_string s) {
  int i, n=N(s);
  char* _s= new char[n+1];
  for (i=0; i<n; i++) _s[i]=s[i];
  _s[i]='\0';
  return _s;
}

mm_string&
operator << (mm_string& s, char x) {
  s->resize (N(s)+ 1);
  s [N(s)-1]= x;
  return s;
}

mm_string&
operator << (mm_string& s1, mm_string s2) {
  register int i, k1= N(s1), k2=N(s2);
  s1->resize (k1+k2);
  for (i=0; i<k2; i++) s1[i+k1]= s2[i];
  return s1;
}

mm_string&
operator << (mm_string& s, int i) {
  return s << as_mm_string (i);
}

mm_string
operator * (mm_string s1, mm_string s2) {
  register int i, n1=N(s1), n2=N(s2);
  mm_string r(n1+n2);
  for (i=0; i<n1; i++) r[i]=s1[i];
  for (i=0; i<n2; i++) r[i+n1]=s2[i];
  return r;
}
